// Copyright (c) 2023 Harry [Majored] [hello@majored.pw]
// MIT License (https://github.com/Majored/rs-async-zip/blob/main/LICENSE)

//! A base runtime-agnostic implementation using `futures`'s IO types.

pub mod read;
pub mod write;
